@@ -3,19 +3,18 @@ import csv
 import re
 from doc2dict import xml2dict, txt2dict, dict2dict
 from doc2dict.mapping import flatten_hierarchy
-from doc2dict import html2dict, visualize_dict, get_title, unnest_dict, pdf2dict
+from doc2dict import html2dict, visualize_dict, get_title, unnest_dict, pdf2dict, flatten_dict
 from ..mapping_dicts.txt_mapping_dicts import dict_10k, dict_10q, dict_8k, dict_13d, dict_13g
 from ..mapping_dicts.xml_mapping_dicts import dict_345
 from ..mapping_dicts.html_mapping_dicts import *
-from selectolax.parser import HTMLParser
-
 from pathlib import Path
 import webbrowser
 from secsgml.utils import bytes_to_str
 import tempfile
 from .tables.tables import Tables
 
-from ..tags.utils import get_cusip_using_regex, get_isin_using_regex, get_figi_using_regex,get_all_tickers, get_full_names,get_full_names_dictionary_lookup
+from ..tags.utils import get_cusip_using_regex, get_isin_using_regex, get_figi_using_regex,get_all_tickers, get_full_names,get_full_names_dictionary_lookup, analyze_lm_sentiment_fragment
+from ..utils.pdf import has_extractable_text
 
 class DataWithTags(dict):
     def __init__(self, data, document):
@@ -29,6 +28,12 @@ class DataWithTags(dict):
             self._tags = Tags(self._document, mode='data')  # New fragment-based behavior
         return self._tags
     
+    @property
+    def similarity(self):
+        if not hasattr(self, '_similarity'):
+            self._similarity = Similarity(self._document, mode='data')
+        return self._similarity
+    
 class TextWithTags(str):
     def __new__(cls, content, document):
         instance = str.__new__(cls, content)
@@ -42,6 +47,12 @@ class TextWithTags(str):
             self._tags = Tags(self._document, mode='text')  # Original behavior
         return self._tags
     
+    @property
+    def similarity(self):
+        if not hasattr(self, '_similarity'):
+            self._similarity = Similarity(self._document, mode='text')
+        return self._similarity
+        
 
 class Tickers:
     def __init__(self, document):
@@ -77,12 +88,11 @@ class Tickers:
         data = self._get_tickers_data()
         return str(data)
     
-class Tags:
+class TextAnalysisBase:
     def __init__(self, document, mode='text'):
         from ..tags.config import _active_dictionaries,_loaded_dictionaries
         self.document = document
         self.mode = mode  # 'text' or 'data'
-        self._tickers = None
         self.dictionaries = {}
         self.processors = {}
         self._text_sources = None
@@ -102,29 +112,9 @@ class Tags:
                 # Original behavior - single text source
                 self._text_sources = [{'id': None, 'text': str(self.document.text)}]
             else:  # mode == 'data'
-                # New behavior - multiple text fragments
-                self._text_sources = []
-                self._extract_text_fragments(self.document.data, '')
+                self._text_sources = [{'id':data_tuple[0],'text':data_tuple[2]} for data_tuple in self.document.data_tuples if data_tuple[1] in ['text','title','textsmall']]
         return self._text_sources
     
-    def _extract_text_fragments(self, data, parent_id=''):
-        """Extract all text fragments with their document IDs from parsed data"""
-        if isinstance(data, dict):
-            for key, value in data.items():
-                if key in ["text", "title"] and isinstance(value, str):
-                    # Use the current dictionary's parent key as the fragment ID
-                    self._text_sources.append({
-                        'id': parent_id,
-                        'text': value
-                    })
-                elif isinstance(value, (dict, list)):
-                    # Pass the current key as the parent_id for the next level
-                    self._extract_text_fragments(value, key)
-        elif isinstance(data, list):
-            for i, item in enumerate(data):
-                if isinstance(item, (dict, list)):
-                    self._extract_text_fragments(item, parent_id)
-    
     def _format_results(self, results, fragment_id):
         """Format results based on mode"""
         if self.mode == 'text':
@@ -133,6 +123,11 @@ class Tags:
         else:
             # New format: (match, fragment_id, start, end)
             return [(match, fragment_id, start, end) for match, start, end in results]
+
+class Tags(TextAnalysisBase):
+    def __init__(self, document, mode='text'):
+        super().__init__(document, mode)
+        self._tickers = None
     
     @property
     def cusips(self):
@@ -205,7 +200,6 @@ class Tags:
         if not hasattr(self, '_persons'):
             self._persons = []
             sources = self._get_text_sources()
-            
             for source in sources:
                 if '8k_2024_persons' in self.processors:
                     results = get_full_names_dictionary_lookup(source['text'], self.processors['8k_2024_persons'])
@@ -218,7 +212,38 @@ class Tags:
                 self._persons.extend(formatted_results)
                     
         return self._persons
-    
+
+class Similarity(TextAnalysisBase):
+    @property
+    def loughran_mcdonald(self):
+        if not hasattr(self, '_loughran_mcdonald'):
+            self._loughran_mcdonald = []
+            sources = self._get_text_sources()
+            
+            if 'loughran_mcdonald' in self.processors:
+                lm_processors = self.processors['loughran_mcdonald']
+                
+                for source in sources:
+                    results = analyze_lm_sentiment_fragment(source['text'], lm_processors)
+                    
+                    if self.mode == 'text':
+                        # Single result for whole document
+                        self._loughran_mcdonald = results
+                        break
+                    else:
+                        # Per-fragment results with fragment_id
+                        fragment_result = {
+                            'fragment_id': source['id'],
+                            **results
+                        }
+                        self._loughran_mcdonald.append(fragment_result)
+            else:
+                # No processors available
+                self._loughran_mcdonald = [] if self.mode == 'data' else {}
+                    
+        return self._loughran_mcdonald
+
+
 class Document:
     def __init__(self, type, content, extension,accession,filing_date,path=None):
         
@@ -240,103 +265,40 @@ class Document:
 
         # this will be filled by parsed
         self._data = None
+        self._data_tuples = None
         self._tables = None
         self._text = None
+        self._markdown = None
+
+        # booleans
+        self._data_bool = self.extension in ('.htm', '.html','.txt')
+
+        # may slow things down?
+        if self.extension == '.pdf':
+            if has_extractable_text(pdf_bytes=self.content):
+                self._data_bool = True
+
+        self._data_tuples_bool = self._data_bool
+        self._text_bool = self._data_bool
+        self._markdown_bool = self._data_bool
+        self._visualize_bool = self._data_bool
+        self._tables_bool = self.extension in ('.xml')
         
 
 
-    #_load_text_content
-    def _preprocess_txt_content(self):
-            self._text = self.content.decode().translate(str.maketrans({
-                '\xa0': ' ', '\u2003': ' ',
-                '\u2018': "'", '\u2019': "'",
-                '\u201c': '"', '\u201d': '"'
-            }))
-
-    # needs work
-    def _preprocess_html_content(self):
-        parser = HTMLParser(self.content,detect_encoding=True,decode_errors='ignore')
-    
-        # Remove hidden elements first
-        hidden_nodes = parser.css('[style*="display: none"], [style*="display:none"], .hidden, .hide, .d-none')
-        for node in hidden_nodes:
-            node.decompose()
-        
-        blocks = {'p', 'div', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'article', 'section', 'li', 'td'}
-        lines = []
-        current_line = []
-        
-        def flush_line():
-            if current_line:
-                # Don't add spaces between adjacent spans
-                lines.append(''.join(current_line))
-                current_line.clear()
-        
-        for node in parser.root.traverse(include_text=True):
-            if node.tag in ('script', 'style', 'css'):
-                continue
-                
-            if node.tag in blocks:
-                flush_line()
-                lines.append('')
-                
-            if node.text_content:
-                text = node.text_content.strip()
-                if text:
-                    if node.tag in blocks:
-                        flush_line()
-                        lines.append(text)
-                        lines.append('')
-                    else:
-                        # Only add space if nodes aren't directly adjacent
-                        if current_line and not current_line[-1].endswith(' '):
-                            if node.prev and node.prev.text_content:
-                                if node.parent != node.prev.parent or node.prev.next != node:
-                                    current_line.append(' ')
-                        current_line.append(text)
-        
-        flush_line()
-        
-        text = '\n'.join(lines)
-        while '\n\n\n' in text:
-            text = text.replace('\n\n\n', '\n\n')
-        
-        self._text = text.translate(str.maketrans({
-            '\xa0': ' ', '\u2003': ' ',
-            '\u2018': "'", '\u2019': "'",
-            '\u201c': '"', '\u201d': '"'
-        }))
-
     def contains_string(self, pattern):
         """Works for select files"""
         if self.extension in ['.htm', '.html', '.txt','.xml']:
             return bool(re.search(pattern, self.content))
         return False
     
-    # Note: this method will be heavily modified in the future
     def parse(self):
         # check if we have already parsed the content
         if self._data:
             return
         
         mapping_dict = None
-        
-        if self.extension == '.txt':
-            content = self.text
-            if self.type in ['10-Q', '10-Q/A']:
-                mapping_dict = dict_10q
-            elif self.type in ['10-K','10-K/A']:
-                mapping_dict = dict_10k
-            elif self.type in ['8-K', '8-K/A']:
-                mapping_dict = dict_8k
-            elif self.type in ['SC 13D', 'SC 13D/A']:
-                mapping_dict = dict_13d
-            elif self.type in ['SC 13G', 'SC 13G/A']:
-                mapping_dict = dict_13g
-            
-            self._data = {}
-            self._data['document'] = dict2dict(txt2dict(content=content, mapping_dict=mapping_dict))
-        elif self.extension in ['.htm', '.html']:
+        if self._data_bool:
             
             if self.type in ['1-K', '1-K/A']:
                 mapping_dict = dict_1kpartii_html
@@ -410,45 +372,73 @@ class Document:
                 mapping_dict = dict_t3_html
             elif self.type in ['NT 10-K', 'NT 10-K/A', 'NT 10-Q', 'NT 10-Q/A', 'NT 20-F', 'NT 20-F/A']:
                 mapping_dict = dict_nt10k_html
+            elif self.type in ['SC 13G', 'SC 13G/A']:
+                mapping_dict = dict_13g
+            elif self.type in ['SC 13D', 'SC 13D/A']:
+                mapping_dict = dict_13d
+            
+            if self.extension in ['.htm','.html']:
+                dct = html2dict(content=self.content, mapping_dict=mapping_dict)
+            elif self.extension in ['.txt']:
+                dct = txt2dict(content=self.content, mapping_dict=mapping_dict)
+            elif self.extension == '.pdf':
+                dct = pdf2dict(content=self.content, mapping_dict=mapping_dict)
+            else:
+                dct = {}
             
-            dct = html2dict(content=self.content, mapping_dict=mapping_dict)
             self._data = dct
         elif self.extension == '.xml':
             if self.type in ['3', '4', '5', '3/A', '4/A', '5/A']:
                 mapping_dict = dict_345
-            
             self._data = xml2dict(content=self.content, mapping_dict=mapping_dict)
-        elif self.extension == '.pdf':
-            self._data = pdf2dict(content=self.content, mapping_dict=mapping_dict)
+
         else:
             pass
 
     @property
     def data(self):
-        if self._data is None:
-            self.parse()
+        if self._data_bool:
+            if self._data is None:
+                self.parse()
 
-        if self._data is None:
-            self._data = {}
-        
-        if not isinstance(self._data, DataWithTags):
-            self._data = DataWithTags(self._data, self)
+            if self._data is None:
+                self._data = {}
+            
+            if not isinstance(self._data, DataWithTags):
+                self._data = DataWithTags(self._data, self)
+        elif self.extension == '.xml':
+            if self._data is None:
+                self.parse()
+
+            if self._data is None:
+                self._data = {}
             
         return self._data
     
+    @property
+    def data_tuples(self):
+        if self._data_bool:
+            if self._data_tuples is None:
+                self._data_tuples = unnest_dict(self.data)
+        return self._data_tuples
+    
     @property
     def text(self):
-        if self._text is None:
-            if self.extension in ['.htm','.html']:
-                self._preprocess_html_content()  # Still sets self._text to plain string
-            elif self.extension == '.txt':
-                self._preprocess_txt_content()   # Still sets self._text to plain string
-            
-            # Convert the plain string to TextWithTags
-            plain_text = self._text
-            self._text = TextWithTags(plain_text, self)
+        if self._text_bool:
+            if self._text is None:
+                text = flatten_dict(tuples_list=self.data_tuples,format='text')
+                self._text = TextWithTags(text, self)
         return self._text
     
+    @property
+    def markdown(self):
+        if self._markdown_bool:
+            if self._markdown is None:
+                self._markdown = flatten_dict(tuples_list=self.data_tuples,format='markdown')
+
+        return self._markdown
+
+    
     def write_json(self, output_filename=None):
         if not self.data:
             self.parse()
@@ -457,19 +447,46 @@ class Document:
             json.dump(self.data, f, indent=2)
 
     def parse_tables(self,must_exist_in_mapping=True):
-        if self.extension != '.xml':
-            self._tables = []
+        """Must exist in mapping means columns must occur in mapping schema."""
+        if self.extension == '.xml':
+            tables = Tables(document_type = self.type, accession=self.accession)
+            tables.parse_tables(data=self.data,must_exist_in_mapping=must_exist_in_mapping)
+            self._tables = tables
+
+        elif self._data_bool:
+            tables = Tables(document_type = self.type, accession=self.accession)
+            data_tuples = self.data_tuples
+            
+            for i, (id, type, content, level) in enumerate(data_tuples):
+                if type == "table" and i > 0:
+                    description = None
+                    
+                    # Look at previous element
+                    prev_id, prev_type, prev_content, prev_level = data_tuples[i-1]
+                    
+                    # Case 1: Same level + text content
+                    if prev_level == level and prev_type in ["text", "textsmall"]:
+                        description = prev_content
+                    
+                    # Case 2: Higher level (lower number) + title
+                    elif prev_level < level and prev_type == "title":
+                        description = prev_content
+                    
+                    # Case 3: No matching description - add table without description
+                    # (description remains None)
+                    
+                    tables.add_table(data=content, description=description, name="extracted_table")
+
+            self._tables = tables
+
         else:
-            # Use the property to trigger parsing if needed
-            data = self.data
-            tables = Tables(document_type = self.type, accession=self.accession, data=data,must_exist_in_mapping=must_exist_in_mapping)
-            self._tables = tables.tables
+            self._tables = []
 
     @property
     def tables(self):
         if self._tables is None:
             self.parse_tables()
-        return self._tables
+        return self._tables.tables
 
 
     def write_csv(self, output_folder):
@@ -497,6 +514,17 @@ class Document:
                     writer.writeheader()
                     writer.writerows(table.data)
 
+    def reset_nlp(self):
+        """Reset all NLP analysis by creating fresh wrapper objects"""
+        # Reset data wrapper
+        if hasattr(self, '_data') and self._data is not None:
+            raw_data = dict(self._data)  # Extract the underlying dict
+            self._data = DataWithTags(raw_data, self)
+        
+        # Reset text wrapper
+        if hasattr(self, '_text') and self._text is not None:
+            raw_text = str(self._text)  # Extract the underlying string
+            self._text = TextWithTags(raw_text, self)
         
     def _document_to_section_text(self, document_data, parent_key=''):
         items = []
@@ -551,73 +579,18 @@ class Document:
             print(f"Cannot open files with extension {self.extension}")
 
     def get_section(self, title=None, title_regex=None,title_class=None, format='dict'):
-        if not self.data:
-            self.parse()
+        if self._data_bool:
+            if not self.data:
+                self.parse()
 
-        result = get_title(self.data,title=title,title_regex=title_regex,title_class=title_class)
-
-        if format == 'text':
-            result = [item[1] for item in result]
-            result = [unnest_dict(item) for item in result]
-
-        return result
-
-   
-   # TODO CHANGE THIS
-    def __iter__(self):
-        # Use the property to trigger parsing if needed
-        document_data = self.data
-
-        # Let's remove XML iterable for now
-
-        # Handle text-based documents
-        if self.extension in ['.txt', '.htm', '.html']:
-            if not document_data:
-                return iter([])
-                
-            # Find highest hierarchy level from mapping dict
-            highest_hierarchy = float('inf')
-            section_type = None
-            
-            if self.type in ['10-K', '10-Q']:
-                mapping_dict = dict_10k if self.type == '10-K' else dict_10q
-            elif self.type == '8-K':
-                mapping_dict = dict_8k
-            elif self.type == 'SC 13D':
-                mapping_dict = dict_13d
-            elif self.type == 'SC 13G':
-                mapping_dict = dict_13g
+            result = get_title(self.data,title=title,title_regex=title_regex,title_class=title_class)
+            if format == 'dict':
+                return [item[1] for item in result]
             else:
-                return iter([])
-                
-            # Find section type with highest hierarchy number
-            highest_hierarchy = -1  # Start at -1 to find highest
-            for mapping in mapping_dict['rules']['mappings']:
-                if mapping.get('hierarchy') is not None:
-                    if mapping['hierarchy'] > highest_hierarchy:
-                        highest_hierarchy = mapping['hierarchy']
-                        section_type = mapping['name']
-                        
-            if not section_type:
-                return iter([])
-                
-            # Extract sections of the identified type
-            def find_sections(data, target_type):
-                sections = []
-                if isinstance(data, dict):
-                    if data.get('type') == target_type:
-                        sections.append({
-                            'item': data.get('text', ''),
-                            'text': flatten_hierarchy(data.get('content', []))
-                        })
-                    for value in data.values():
-                        if isinstance(value, (dict, list)):
-                            sections.extend(find_sections(value, target_type))
-                elif isinstance(data, list):
-                    for item in data:
-                        sections.extend(find_sections(item, target_type))
-                return sections
-                
-            return iter(find_sections(document_data, section_type))
-            
-        return iter([])
\ No newline at end of file
+                return [flatten_dict(item[1],format) for item in result]
+
+    # TODO
+    def get_tables(self,description_regex=None,name=None):
+        # make sure tables is initialized
+        self.tables
+        return self._tables.get_tables(description_regex=description_regex, name=name)
\ No newline at end of file
