@@ -1,8 +1,14 @@
-from ..utils.dictionaries import download_dictionary, load_dictionary
+from .dictionaries import download_dictionary, load_dictionary
 
 _active_dictionaries = []
 _loaded_dictionaries = {}
 
+def clear_dictionaries():
+    """Remove all active dictionaries"""
+    global _active_dictionaries, _loaded_dictionaries
+    _active_dictionaries = []
+    _loaded_dictionaries = {}
+
 def set_dictionaries(dictionaries, overwrite=False):
     """Set active dictionaries and load them into memory"""
     global _active_dictionaries, _loaded_dictionaries
@@ -26,6 +32,14 @@ def set_dictionaries(dictionaries, overwrite=False):
                 'data': raw_data,
                 'processor': processor
             }
+        elif dict_name == 'loughran_mcdonald':
+            from .utils import create_lm_processors
+            processors = create_lm_processors(raw_data)
+            
+            _loaded_dictionaries[dict_name] = {
+                'data': raw_data,
+                'processor': processors 
+            }
         else:
             _loaded_dictionaries[dict_name] = {
                 'data': raw_data,
